from typing import List
from pygazpar.enum import NatureReleve, QualificationReleve, StatusReleve


class RelevesType:
    """Class representing a Releves from API"""
    def __init__(self,
                dateDebutReleve:str,
                dateFinReleve:str,
                indexDebut:int,
                indexFin:int,
                volumeBrutConsomme:float,
                energieConsomme:float,
                natureReleve:str|NatureReleve,
                qualificationReleve:QualificationReleve|str,
                journeeGaziere:str| None=None,
                pcs:str|int|float|None=None,
                volumeConverti:int|float|None=None,
                pta:str|int|float|None=None,
                status:StatusReleve|str|None=None,
                coeffConversion:float|None=None,
                frequenceReleve:str|None=None,
                temperature:str|float=None,
                frequence:str|None=None):
        self.dateDebutReleve = dateDebutReleve
        self.dateFinReleve = dateFinReleve
        self.journeeGaziere = journeeGaziere
        self.indexDebut = indexDebut
        self.indexFin = indexFin
        self.volumeBrutConsomme = volumeBrutConsomme
        self.energieConsomme = energieConsomme
        self.pcs = pcs
        self.volumeConverti = volumeConverti
        self.pta = pta
        if(natureReleve is None):
            self.natureReleve = None
        else:   
            self.natureReleve =  NatureReleve(natureReleve)
        if(qualificationReleve is None):
            self.qualificationReleve = None
        else:   
            self.qualificationReleve =  QualificationReleve(qualificationReleve)
        if(status is None):
            self.status = None
        else:   
            self.status =  StatusReleve(status)
        self.coeffConversion = coeffConversion
        self.frequenceReleve = frequenceReleve
        self.temperature = temperature
        self.frequence = frequence

class ConsommationType:
    """Class representing a Result consommation send by the API"""
    def __init__(self,
                 idPce:str,
                 releves:List[RelevesType],
                 frequence:str|None):
        self.idPce = idPce
        relevesArray=[]
        for element in releves:
            if not isinstance(element, RelevesType):
                relevesArray.append(RelevesType(**element))
            else:
                relevesArray.append(element)
        self.releves = relevesArray
        self.frequence = frequence

# ------------------------------------------------------------------------------------------------------------


import sys
import os
import dotenv
import json
import pygazpar


def main():

    # Load environment variables GRDF_USERNAME, GRDF_PASSWORD, PCE_IDENTIFIER from .env file.
    dotenv.load_dotenv()

    username = os.environ["GRDF_USERNAME"]
    password = os.environ["GRDF_PASSWORD"]
    pceIdentifier = os.environ["PCE_IDENTIFIER"]

    client = pygazpar.Client(pygazpar.JsonWebDataSource(
        username=username,
        password=password)
    )

    data = client.loadSince(pceIdentifier=pceIdentifier,
                            lastNDays=60,
                            frequencies=[pygazpar.Frequency.DAILY, pygazpar.Frequency.MONTHLY])

    print(json.dumps(data, indent=2))


if __name__ == '__main__':
    sys.exit(main())



"""Support for Helper."""
from typing import Any
import socket
import aiohttp
import async_timeout
from pygazpar.exceptions import ClientAuthenticationError, ClientCommunicationError,ClientError

async def _api_wrapper(
    session:aiohttp.ClientSession,
    method: str,
    url: str,
    data: dict | None = None,
    headers: dict | None = None,
    params: dict | None = None,
) -> Any:
    """Get information from the API."""
    try:
        async with async_timeout.timeout(10):
            response = await session.request(
                method=method,
                url=url,
                headers=headers,
                json=data,
                params=params,
            )
            _verify_response_or_raise(response)
            return response

    except TimeoutError as exception:
        msg = f"Timeout error fetching information - {exception}"
        raise ClientCommunicationError(
            msg,
        ) from exception
    except (aiohttp.ClientError, socket.gaierror) as exception:
        msg = f"Error fetching information - {exception}"
        raise ClientCommunicationError(
            msg,
        ) from exception
    except Exception as exception:  # pylint: disable=broad-except
        msg = f"Something really wrong happened! - {exception}"
        raise ClientError(
            msg,
        ) from exception
def _verify_response_or_raise(response: aiohttp.ClientResponse) -> None:
    """Verify that the response is valid."""
    if response.status in (401, 403):
        msg = "Invalid credentials"
        raise ClientAuthenticationError(
            msg,
        )
    response.raise_for_status()



"""Support for Enum."""
from enum import Enum


# ------------------------------------------------------------------------------------------------------------
class PropertyName(Enum):
    '''Get the name of the field'''
    TIME_PERIOD = "time_period"
    DATE_DEBUT = "dateDebutReleve"
    DATE_FIN = "dateFinReleve"

    JOURNEE_GAZIERE = "journeeGaziere"
    START_INDEX = "indexDebut"
    END_INDEX = "indexFin"
    VOLUME = "volumeBrutConsomme"
    ENERGY = "energieConsomme"
    PCS = "pcs"
    VOLUME_CONVERTI = "volumeConverti"
    PTA = "pta"
    NATURE = "natureReleve"
    QUALIFICATION = "qualificationReleve"
    STATUS = "status"
    FREQUENCE_RELEVE= "frequenceReleve"
    CONVERTER_FACTOR = "coeffConversion"
    TEMPERATURE = "temperature"
    FREQUENCE= "frequence"
    TIMESTAMP = "timestamp"                
    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()


# ------------------------------------------------------------------------------------------------------------
class Frequency(Enum):
    '''Get frequency'''
    HOURLY = "hourly"
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    YEARLY = "yearly"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()
class ConsommationRole(str,Enum):
    '''Get type conso for API'''
    INFORMATIVES = 'informatives'
    PUBLIEES = 'publiees'
class NatureReleve(str,Enum):
    '''Get nature type for releve'''
    PUBLIEES = 'Publiée'
    INFORMATIVES = 'Informative Journalier'
class QualificationReleve(str,Enum):
    '''Get qualification type for releve'''
    ESTIME='Estimé'
    CORRIGE='Corrigé'
    MESURE='Mesuré'
    ABSENT='Absence de Données'
class StatusReleve(str,Enum):
    '''Get status type for releve'''
    PROVISOIRE='Provisoire'
    DEFINITIVE='Définitive'



"""Support for Frequency Converter."""

from typing import List, Dict, Any, cast
import pandas as pd
from pygazpar.types.RelevesResultType import RelevesResultType
from pygazpar.enum import NatureReleve, QualificationReleve

# ------------------------------------------------------------------------------------------------------------
class FrequencyConverter:
    """Class for Convert daily data to other frequency."""
    INPUT_DATE_FORMAT = "%Y-%m-%d"
    OUTPUT_DATE_FORMAT = "%d/%m/%Y"
    OUTPUT2_DATE_FORMAT = "%Y-%m-%d "

    MONTHS = [
        "Janvier",
        "Février",
        "Mars",
        "Avril",
        "Mai",
        "Juin",
        "Juillet",
        "Août",
        "Septembre",
        "Octobre",
        "Novembre",
        "Décembre"
    ]

    # ------------------------------------------------------
    @staticmethod
    def compute_hourly(daily: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Compute hourly data."""

        return []

    # ------------------------------------------------------
    @staticmethod
    def compute_daily(daily: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Compute hourly data."""
        return daily

    # ------------------------------------------------------
    @staticmethod
    def compute_weekly(daily: List[RelevesResultType]) -> List[RelevesResultType]:
        """Compute Weekly data."""
        convertDaily = [ob.__dict__ for ob in daily]
        df = pd.DataFrame(convertDaily)

        # Trimming head and trailing spaces and convert to datetime.
        #df["date_time"] = pd.to_datetime(df["time_period"].str.strip(), format=FrequencyConverter.OUTPUT_DATE_FORMAT)
        df["journeeGaziere"] = pd.to_datetime(df["journeeGaziere"], format=FrequencyConverter.INPUT_DATE_FORMAT)

        # Get the first day of week.

        df["dateDebutReleve"] = pd.to_datetime(df["journeeGaziere"].dt.strftime("%W %Y 1T06:00:00"), format="%W %Y %wT%H:%M:%S")
        df["dateDebutReleve"]= df["dateDebutReleve"].dt.tz_localize("Europe/Paris")

        df["dateDebutWeek"] = pd.to_datetime(df["journeeGaziere"].dt.strftime("%W %Y 1"), format="%W %Y %w")
        # Get the last day of week.

        df["dateFinReleve"] = pd.to_datetime(df["journeeGaziere"].dt.strftime("%W %Y 0T06:00:00"), format="%W %Y %wT%H:%M:%S")
        df["dateFinReleve"]= df["dateFinReleve"].dt.tz_localize("Europe/Paris")
        df["dateFinReleve"]= df["dateFinReleve"]+pd.Timedelta(days=1)

        df["dateFinWeek"] = pd.to_datetime(df["journeeGaziere"].dt.strftime("%W %Y 0"), format="%W %Y %w")

        # Reformat the time period.
        df["time_period"] = "Du " + df["dateDebutWeek"].dt.strftime(FrequencyConverter.OUTPUT_DATE_FORMAT).astype(str) + " au " + df["dateFinWeek"].dt.strftime(FrequencyConverter.OUTPUT_DATE_FORMAT).astype(str)

        # Aggregate rows by month_year.
        df = df[["dateDebutReleve","dateFinReleve", "time_period", "indexDebut", "indexFin", "volumeBrutConsomme", "energieConsomme","temperature", "timestamp"]].groupby("time_period").agg(dateDebutReleve=('dateDebutReleve', 'min'),dateFinReleve=('dateFinReleve', 'max'), indexDebut=('indexDebut', 'min'), indexFin=('indexFin', 'max'), volumeBrutConsomme=('volumeBrutConsomme', 'sum'), energieConsomme=('energieConsomme', 'sum'),  temperature=('temperature', 'mean'),timestamp=('timestamp', 'min'), count=('energieConsomme', 'count')).reset_index()

        # Sort rows by month ascending.
        df = df.sort_values(by=['dateDebutReleve'])

        # Select rows where we have a full week (7 days) except for the current week.
        df = pd.concat([df[(df["count"] >= 7)], df.tail(1)[df.tail(1)["count"] < 7]])
        
        df['dateDebutReleve']=df['dateDebutReleve'].apply(FrequencyConverter.convert_datetime_iso_string)
        df['dateFinReleve']=df['dateFinReleve'].apply(FrequencyConverter.convert_datetime_iso_string)

        # Select target columns.
        df = df[["time_period","dateDebutReleve","dateFinReleve", "indexDebut", "indexFin", "volumeBrutConsomme", "energieConsomme","temperature", "timestamp"]]
        res = cast(List[Dict[str, Any]], df.to_dict('records'))
        result = [RelevesResultType(**dict(item, **{'natureReleve':NatureReleve.INFORMATIVES.value,'qualificationReleve':QualificationReleve.ESTIME.value})) for item in res]

        return result
    @staticmethod
    def convert_datetime_iso_string(x) -> str:
        """Compute Datetime to iso string data."""
        return pd.Timestamp(x).isoformat()
    # ------------------------------------------------------
    @staticmethod
    def compute_monthly(daily: List[RelevesResultType]) -> List[RelevesResultType]:
        """Compute Monthly data."""
        convertDaily = [ob.__dict__ for ob in daily]
        df = pd.DataFrame(convertDaily)

        # Trimming head and trailing spaces and convert to datetime.
        df["journeeGaziere"] = pd.to_datetime(df["journeeGaziere"], format=FrequencyConverter.INPUT_DATE_FORMAT)
        
        df["dateDebutReleve"] = pd.to_datetime(df["journeeGaziere"].dt.strftime("%Y %m 01T06:00:00"), format="%Y %m %dT%H:%M:%S")
        df["dateDebutReleve"]= df["dateDebutReleve"].dt.tz_localize("Europe/Paris")
        
        df["dateFinReleve"] = pd.to_datetime(df["journeeGaziere"].dt.strftime("%Y %m 01T06:00:00"), format="%Y %m %dT%H:%M:%S")
        df["dateFinReleve"]= df["dateFinReleve"].dt.tz_localize("Europe/Paris")
        df["dateFinReleve"]= df["dateFinReleve"]+pd.DateOffset(months=1)

        # Get the corresponding month-year.
        df["month_year"] = df["journeeGaziere"].apply(lambda x: FrequencyConverter.MONTHS[x.month - 1]).astype(str) + " " + df["journeeGaziere"].dt.strftime("%Y").astype(str)

        # Aggregate rows by month_year.
        #df = df[["date_time", "month_year", "start_index_m3", "end_index_m3", "volume_m3", "energy_kwh", "timestamp"]].groupby("month_year").agg(first_day_of_month=('date_time', 'min'), start_index_m3=('start_index_m3', 'min'), end_index_m3=('end_index_m3', 'max'), volume_m3=('volume_m3', 'sum'), energy_kwh=('energy_kwh', 'sum'), timestamp=('timestamp', 'min'), count=('energy_kwh', 'count')).reset_index()
        df = df[["dateDebutReleve","dateFinReleve", "month_year", "indexDebut", "indexFin", "volumeBrutConsomme", "energieConsomme","temperature", "timestamp"]].groupby("month_year").agg(dateDebutReleve=('dateDebutReleve', 'min'),dateFinReleve=('dateFinReleve', 'max'), indexDebut=('indexDebut', 'min'), indexFin=('indexFin', 'max'), volumeBrutConsomme=('volumeBrutConsomme', 'sum'), energieConsomme=('energieConsomme', 'sum'),  temperature=('temperature', 'mean'),timestamp=('timestamp', 'min'), count=('energieConsomme', 'count')).reset_index()

        # Sort rows by month ascending.
        df = df.sort_values(by=['dateDebutReleve'])

        # Select rows where we have a full month (more than 27 days) except for the current month.
        df = pd.concat([df[(df["count"] >= 28)], df.tail(1)[df.tail(1)["count"] < 28]])

        # Rename columns for their target names.
        df = df.rename(columns={"month_year": "time_period"})

        # Select target columns.
        df = df[["time_period","dateDebutReleve","dateFinReleve", "indexDebut", "indexFin", "volumeBrutConsomme", "energieConsomme","temperature", "timestamp"]]
        
        df['dateDebutReleve']=df['dateDebutReleve'].apply(FrequencyConverter.convert_datetime_iso_string)
        df['dateFinReleve']=df['dateFinReleve'].apply(FrequencyConverter.convert_datetime_iso_string)
        res = cast(List[Dict[str, Any]], df.to_dict('records'))
        result = [RelevesResultType(**dict(item, **{'natureReleve':NatureReleve.INFORMATIVES.value,'qualificationReleve':QualificationReleve.ESTIME.value})) for item in res]

        return result

    # ------------------------------------------------------
    @staticmethod
    def compute_yearly(daily: List[RelevesResultType]) -> List[RelevesResultType]:
        """Compute Yearly data."""
        convertDaily = [ob.__dict__ for ob in daily]
        df = pd.DataFrame(convertDaily)

        # Trimming head and trailing spaces and convert to datetime.
        df["journeeGaziere"] = pd.to_datetime(df["journeeGaziere"], format=FrequencyConverter.INPUT_DATE_FORMAT)

         
        df["dateDebutReleve"] = pd.to_datetime(df["journeeGaziere"].dt.strftime("%Y 01 01T06:00:00"), format="%Y %m %dT%H:%M:%S")
        df["dateDebutReleve"]= df["dateDebutReleve"].dt.tz_localize("Europe/Paris")
        
        df["dateFinReleve"] = pd.to_datetime(df["journeeGaziere"].dt.strftime("%Y 12 01T06:00:00"), format="%Y %m %dT%H:%M:%S")
        df["dateFinReleve"]= df["dateFinReleve"].dt.tz_localize("Europe/Paris")
        df["dateFinReleve"]= df["dateFinReleve"]+pd.DateOffset(months=1)
        # Get the corresponding year.
        df["year"] = df["journeeGaziere"].dt.strftime("%Y")

        # Aggregate rows by month_year.
        df = df[["year", "dateDebutReleve","dateFinReleve", "indexDebut", "indexFin","volumeBrutConsomme", "energieConsomme","temperature" "timestamp"]].groupby("year").agg(start_index_m3=('indexDebut', 'min'), end_index_m3=('indexFin', 'max'), volume_m3=('volumeBrutConsomme', 'sum'), energy_kwh=('energieConsomme', 'sum'),  temperature=('temperature', 'mean'),timestamp=('timestamp', 'min'), count=('energieConsomme', 'count')).reset_index()

        # Sort rows by month ascending.
        df = df.sort_values(by=['year'])

        # Select rows where we have almost a full year (more than 360) except for the current year.
        df = pd.concat([df[(df["count"] >= 360)], df.tail(1)[df.tail(1)["count"] < 360]])

        # Rename columns for their target names.
        df = df.rename(columns={"year": "time_period"})

        # Select target columns.
        df = df[["time_period",  "dateDebutReleve","dateFinReleve", "indexDebut", "indexFin","volumeBrutConsomme", "energieConsomme","temperature", "timestamp"]]
        
        df['dateDebutReleve']=df['dateDebutReleve'].apply(FrequencyConverter.convert_datetime_iso_string)
        df['dateFinReleve']=df['dateFinReleve'].apply(FrequencyConverter.convert_datetime_iso_string)
        res = cast(List[Dict[str, Any]], df.to_dict('records'))
        result = [RelevesResultType(**dict(item, **{'natureReleve':NatureReleve.INFORMATIVES.value,'qualificationReleve':QualificationReleve.ESTIME.value})) for item in res]

        return result




from pygazpar.types.ContratType import ContratPce

class TechniquePce:
    """Class representing a Technical PCE  from API"""
    def __init__(self, 
                 calibre:str|None, 
                 numeroMatricule:str,
                 codeDebit:str,
                 frequence:str,
                 idPitd:str,
                 libellePitd:str,
                 clientSensibleMig:str,
                 proprieteCompteur:str,
                 proprieteDetendeur:str|None,
                 convertisseur:str|None,
                 proprieteEnregistreur:str|None,
                 roues:int,
                 debit:str,
                 pressionAval:str|None,
                 numeroRue:str|None,
                 nomRue:str|None,
                 complementAdresse:str|None,
                 codePostal:str|None,
                 codeInseeCommune:str|None,
                 commune:str|None,
                 situationCompteur:str|None,
                 accessibiliteCompteur:str|None,
                 reperageRobinetGaz:str|None,
                 numeroSerie:str|None,
                 etatCompteur:str|None,
                 codeEtatTechniquePce:str|None,
                 libelleEtatTechniquePce:str|None,
                 telereleve:bool,
                 codeEtatCommunication: str|None,
                 libelleEtatCommunication:str|None,
                 codeNatureGaz:str|None,
                 libelleNatureGaz: str|None,

                 ):
        self.calibre = calibre
        self.numeroMatricule = numeroMatricule
        self.codeDebit = codeDebit
        self.frequence = frequence
        self.idPitd = idPitd
        self.libellePitd = libellePitd
        self.clientSensibleMig = clientSensibleMig
        self.proprieteCompteur = proprieteCompteur
        self.proprieteDetendeur = proprieteDetendeur
        self.convertisseur = convertisseur
        self.proprieteEnregistreur = proprieteEnregistreur
        self.roues = roues
        self.debit = debit
        self.pressionAval = pressionAval
        self.numeroRue = numeroRue
        self.nomRue = nomRue
        self.complementAdresse = complementAdresse
        self.codePostal = codePostal
        self.codeInseeCommune = codeInseeCommune    
        self.commune = commune
        self.situationCompteur = situationCompteur
        self.accessibiliteCompteur = accessibiliteCompteur
        self.reperageRobinetGaz = reperageRobinetGaz
        self.numeroSerie = numeroSerie
        self.etatCompteur = etatCompteur
        self.codeEtatTechniquePce = codeEtatTechniquePce
        self.libelleEtatTechniquePce = libelleEtatTechniquePce
        self.telereleve = telereleve
        self.codeEtatCommunication = codeEtatCommunication
        self.libelleEtatCommunication = libelleEtatCommunication
        self.codeNatureGaz = codeNatureGaz
        self.libelleNatureGaz = libelleNatureGaz

class DetailsPce:
    """Class representing a PCE details from API"""
    def __init__(self, 
                 technique:TechniquePce, 
                 contrat:ContratPce,
                 statutRestitutionTechnique:str|None,
                 statutRestitutionContrat:str|None,
                 
                 ):
        if not isinstance(technique, TechniquePce) and isinstance(technique, dict):
            self.technique=TechniquePce(**technique)
        else:
            self.technique = technique
        if not isinstance(contrat, ContratPce) and isinstance(contrat, dict):
            self.contrat=ContratPce(**contrat)
        else:
            self.contrat = contrat
        self.statutRestitutionTechnique = statutRestitutionTechnique
        self.statutRestitutionContrat = statutRestitutionContrat


import json
from pygazpar.types.ConsommationType import RelevesType
from pygazpar.enum import NatureReleve, QualificationReleve, StatusReleve


class RelevesResultType(RelevesType):
    """Class representing a result consommation"""
    def __init__(self,
                time_period:str,
                timestamp:str,
                releves:RelevesType|None=None,
                temperature:str|float|None=None,
                dateDebutReleve:str|None=None,
                dateFinReleve:str|None=None,
                journeeGaziere:str|None=None,
                indexDebut:int|None=None,
                indexFin:int|None=None,
                volumeBrutConsomme:float|None=None,
                energieConsomme:float|None=None,
                pcs:str|int|float|None=None,
                volumeConverti:int|float|None=None,
                pta:str|int|float|None=None,
                natureReleve:NatureReleve|str|None=None,
                qualificationReleve:QualificationReleve|str|None=None,
                status:StatusReleve|str|None=None,
                coeffConversion:float|None=None,
                frequenceReleve:str|None=None,
                frequence:str|None=None
                 ):
        if(releves is not None ):
            if(temperature is None):
                temperature=releves.temperature
            super().__init__(dateDebutReleve=releves.dateDebutReleve,
                    dateFinReleve=releves.dateFinReleve,
                    journeeGaziere=releves.journeeGaziere,
                    indexDebut=releves.indexDebut,
                    indexFin=releves.indexFin,
                    volumeBrutConsomme=releves.volumeBrutConsomme,
                    energieConsomme=releves.energieConsomme,
                    pcs=releves.pcs,
                    volumeConverti=releves.volumeConverti,
                    pta=releves.pta,
                    natureReleve=releves.natureReleve,
                    qualificationReleve=releves.qualificationReleve,
                    status=releves.status,
                    coeffConversion=releves.coeffConversion,
                    frequenceReleve=releves.frequenceReleve,
                    temperature=temperature,
                    frequence=releves.frequence)
        else :
            super().__init__(dateDebutReleve=dateDebutReleve,
                    dateFinReleve=dateFinReleve,
                    journeeGaziere=journeeGaziere,
                    indexDebut=indexDebut,
                    indexFin=indexFin,
                    volumeBrutConsomme=volumeBrutConsomme,
                    energieConsomme=energieConsomme,
                    pcs=pcs,
                    volumeConverti=volumeConverti,
                    pta=pta,
                    natureReleve=natureReleve,
                    qualificationReleve=qualificationReleve,
                    status= status,
                    coeffConversion=coeffConversion,
                    frequenceReleve=frequenceReleve,
                    temperature=temperature,
                    frequence=frequence)

        self.time_period = time_period
        self.timestamp = timestamp
    def toJSON(self):
        return json.dumps(
            self,
            default=lambda o: o.__dict__, 
            sort_keys=True,
            indent=4)



"""Support for Consommation Methods."""
from __future__ import annotations
from typing import  Dict, Any
import aiohttp
from pygazpar.helpers import _api_wrapper
from pygazpar.types.ConsommationType import ConsommationType
from pygazpar.enum import ConsommationRole,Frequency
from .exceptions import ClientError

BASE_URL="https://monespace.grdf.fr/api/e-conso/pce/consommation/"
class  GazparConsommation:
     '''Get the consommation JSON or File from the API'''
     # ------------------------------------------------------
     def __init__(self, session: aiohttp.ClientSession):
        self._session = session
     # ------------------------------------------------------
     async def get_consommation(self,pce:str,date_debut:str,date_fin:str,type_conso:ConsommationRole) -> ConsommationType:
          '''Get the consommation from the API'''
          response=await _api_wrapper(
          session=self._session,
          method="get",
          url=BASE_URL+type_conso.value,
          headers={"Content-type": "application/json","X-Requested-With": "XMLHttpRequest"},
          params={"dateDebut":date_debut,"dateFin":date_fin,"pceList[0]":pce}
          )
          if response.content_type=="application/json":
               responsejson=await response.json()
          else: 
               raise ClientError("Invalid response from server")
          return ConsommationType(**responsejson[pce])
     # ------------------------------------------------------
     async def get_consommation_file(self,pce:str,date_debut:str,date_fin:str,type_conso:ConsommationRole,frequency:Frequency) -> Dict[str, Any]:
          '''Get the consommation file from the API'''
          response=await _api_wrapper(
          session=self._session,
          method="get",
          url=BASE_URL+type_conso.value+"/telecharger",
          headers={"Content-type": "application/json","X-Requested-With": "XMLHttpRequest"},
          params={"dateDebut":date_debut,"dateFin":date_fin,"pceList[0]":pce,"frequence":frequency.value}
          )
          if response.content_type=="text/html":
               raise ClientError("Invalid response from server")
          else:
               filename = response.headers["Content-Disposition"].split("filename=")[1]
               filecontent = await response.content()
               return {"filename":filename,"content":filecontent}



"""Support for API Methods."""
from typing import List, Optional
import logging
from datetime import date, timedelta
from pygazpar.enum import Frequency
from pygazpar.datasource import IDataSource, MeterReadingsByFrequency

DEFAULT_LAST_N_DAYS = 365
Logger = logging.getLogger(__name__)


# ---------------------------------------------------------
class Client:
    '''Get the API Client'''
    # ------------------------------------------------------
    def __init__(self, datasource: IDataSource):
        self.__datasource = datasource

    async def async_login(self):
        '''Try to log in'''
        try:
            res = await self.__datasource.login()
        except Exception:
            Logger.error("An unexpected error occured while loading the data", exc_info=True)
            raise
        return res
    # ------------------------------------------------------
    async def load_since(self, pce_identifier: str, last_n_days: int = DEFAULT_LAST_N_DAYS,
                        frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:
        '''Load data since last N days'''
        try:
            end_date = date.today()
            start_date = end_date + timedelta(days=-last_n_days)
            res = await self.load_date_range(pce_identifier, start_date, end_date, frequencies)
        except Exception:
            Logger.error("An unexpected error occured while loading the data", exc_info=True)
            raise

        return res

    # ------------------------------------------------------
    async def load_date_range(self, pce_identifier: str, start_date: date, end_date: date,
                            frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:
        '''Load data since two date'''
        Logger.debug("Start loading the data...")
        try:
            res = await self.__datasource.load(pce_identifier, start_date, end_date, frequencies)
            Logger.debug("The data load terminates normally")
        except Exception:
            Logger.error("An unexpected error occured while loading the data", exc_info=True)
            raise

        return res
    async def load_list_pce(self):
        '''Load data since last N days'''
        try:
            res = await self.__datasource.list_pce()
        except Exception:
            Logger.error("An unexpected error occured while loading the data", exc_info=True)
            raise
        return res



"""Support for PCE Methods."""
from __future__ import annotations
from typing import List
import aiohttp
from pygazpar.types.PceType import PceType
from .helpers import _api_wrapper
from .exceptions import ClientError
BASE_URL="https://monespace.grdf.fr/api/e-conso/pce"
class  GazparPCE:
    """ Class PCE data from the API."""
     # ------------------------------------------------------
    def __init__(self, session: aiohttp.ClientSession):
        self._session = session
    async def get_list_pce(self) -> List[PceType]:
          """ Get all PCE from an account."""
          response=await _api_wrapper(
          session=self._session,
          method="get",
          url=BASE_URL,
          headers={"Content-type": "application/json","X-Requested-With": "XMLHttpRequest"},
          )
          results_pce=[]
          if response.content_type=="application/json":
               responsejson=await response.json()
          else:
               raise ClientError("Invalid response from server")
          for item in responsejson:
               results_pce.append(PceType(**item))
          return results_pce
    async def get_pce_details(self,pce:str) -> PceType:
        """ Get PCE details."""
        response=await _api_wrapper(
        session=self._session,
        method="get",
        url=BASE_URL+"/"+pce+"/details",
        headers={"Content-type": "application/json","X-Requested-With": "XMLHttpRequest"},
        )
        if response.content_type=="application/json":
             responsejson=await response.json()
        else:
             raise ClientError("Invalid response from server")
        return PceType(**responsejson)
    async def get_pce_meteo(self,pce:str,date_fin:str,nb_jours:int) -> any:
        """ Get PCE meteo temp data."""
        response=await _api_wrapper(
        session=self._session,
        method="get",
        url=BASE_URL+"/"+pce+"/meteo",
        headers={"Content-type": "application/json","X-Requested-With": "XMLHttpRequest"},
        params={"dateFinPeriode":date_fin,"nbJours":nb_jours}
        )
        if response.content_type=="application/json":
             return await response.json()
        else:
             raise ClientError("Invalid response from server")




class ContratPce:
    """Contract class from API.

        Class represent a contract

        Attributes:
        ----------
        tarifAcheminement: str
        Ce Tarif d'acheminement gaz du titulaire est choisi par le fournisseur en fonction d'une estimation du volume de gaz consommé.
            - T1 : Consommation inférieure à 6000 kWh / an.
            - T2 : Consommation entre 6000 kWh / an et 300 MWh / an.
            - T3 : Consommation entre 300 MWh / an et 5 GWh / an (essentiellement des entreprises tertiaires ou industrielles de taille moyenne).
            - T4 : Consommation supérieure à 5 GWh / an (grands sites industriels directement rattachés au réseau de distribution de gaz naturel).
            - TP : Tarif de Proximité, il s’agit d’une option tarifaire concernant des clients finaux ayant la possibilité règlementaire de se raccorder au réseau de transport.
            - TB : Tarif Biométhane, il s'agit d'une option tarifaire concernant les producteurs de biométhane
            
        carActuelle: int
        Consommation Annuelle de Référence (CAR) en kWh :Quantité de gaz estimée consommée sur une année, dans des conditions climatiques moyennes.La CAR d’une année N s’appliquent du 1er avril de l’année N au 31 mars de l’année N+1. Elle est mise à jour de manière systématique par le Gestionnaire de Réseau de Distribution (GRD).
        carFuture: int 
        Prochaine Consommation Annuelle de Référence (CAR) en kWh
        profilTypeFutur: str
        Prochain Profil Type
        cja: str|None
        Capacité Journalière d’Acheminement (CJA) en kWh/jour :Quantité maximale d’énergie que le distributeur s'engage à acheminer chaque jour en un point de livraison.Elle se compose d’une souscription annuelle à laquelle peut s’ajouter une souscription mensuelle supplémentaire et/ou une souscription journalière supplémentaire. Ce type de donnée est applicable seulement aux compteurs JJ et pour les clients de tarif T4 ou TP (Tarif de Proximité).
        cjaMensuelle: str|None
        Souscription mensuelle supplémentaire
        cjaJournaliere: str|None
        Souscription journalière supplémentaire
        idCad: str
        nomTitulaire: str|None
        Nom du titulaire du contrat de fourniture d'énergie
        raisonSocialeTitulaire: str|None
        Raison sociale du titulaire (si applicable)
        numeroSiretTitulaire: str|None
        Numéro SIRET du titulaire (si applicable)
        dateMes: str|None
        Date de la Mise En Service (MES) du PCE correspondante au titulaire actif
        dateMhs: str|None
        Date de la Mise Hors Service (MHS) du PCE correspondante au titulaire actif
        statutContractuel: str
        Statut du contrat
        consommationJournalierePlafond: str|None
        Plafond de consommation journalière
        modulationN1: str|None
        Modulation de Stockage Année N-1
        modulationN2: str|None
        Modulation de Stockage Année N-2
        modulationN3: str|None
        Modulation de Stockage Année N-3
        modulationN4: str|None
        Modulation de Stockage Année N-4
        assiette: str|None
        Assiette de compensation de stockage. Moyenne des 2 modulation plus petite.
        fournisseur: str
        Fournisseur d'énergie souscrit
        profil: str
        Le Profil Type (est attribué par le Distributeur) caractérise la répartition de la CAR d’un PCE tout au long de l’année.
        Il est notamment utilisé entre deux relevés pour estimer les quantités journalières d'un PCE.Ce dernier est déterminé automatiquement par le système d’informations de GRDF à partir de la CAR saisie par le fournisseur (puis chaque année à partir de la CAR recalculée par GRDF).
        Dix profils types permettent de définir les usages de consommations du gaz naturel :
        - P000 : Client PCE forfait cuisine
        - P011 : Client Gazpar (compteur 1M) ou à relevé semestriel avec une CAR inférieure à 6000 kWh/an (compteur 6M)
        - P012 : Client Gazpar (compteur 1M) ou à relevé semestriel avec une CAR supérieure ou égale à 6000 kWh/an (compteur 6M)
        - P013 : Client à relevé mensuel (compteur MM) ou journalier (compteur JJ) avec une part hiver corrigée moyenne inférieure ou égale à 39%
        - P014 : Client à relevé mensuel (compteur MM) ou journalier (compteur JJ) avec une part hiver corrigée moyenne entre 39% et 50%
        - P015 : Client à relevé mensuel (compteur MM) ou journalier (compteur JJ) avec une part hiver corrigée moyenne entre 50% et 58%
        - P016 : Client à relevé mensuel (compteur MM) ou journalier (compteur JJ) avec une part hiver corrigée moyenne entre 58% et 69%
        - P017 : Client à relevé mensuel (compteur MM) ou journalier (compteur JJ) avec une part hiver corrigée moyenne entre 69% et 75%
        - P018 : Client à relevé mensuel (compteur MM) ou journalier (compteur JJ) avec une part hiver corrigée moyenne entre 75% et 81%
        - P019 : Client à relevé mensuel (compteur MM) ou journalier (compteur JJ) avec une part hiver corrigée moyenne strictement supérieure à 81%
        Le Profil Type d’une année N s’appliquent du 1er avril de l’année N au 31 mars de l’année N+1.
        Il est mis à jour de manière systématique par le Gestionnaire de Réseau de Distribution (GRD), une fois par an à date fixe, sauf en cas d'évènement spécifique :
         - Premières Mise En Service (MES)
         - Changement de fournisseur avec changement de tarif ou de fréquence de relevé
         - Changement de données tarifaires
         - Corrections d’une erreur manifeste (CAR et/ou profil aberrant)
        
        dateDebutProfil: str
        Date de début de validité du Profil Type actuel
        dateFinProfil: str
        Date de fin de validité du Profil Type actuel    

        """
    def __init__(self, 
                 tarifAcheminement:str, 
                 carActuelle:int,
                 carFuture:int,
                 profilTypeFutur:str,
                 cja:str|None,
                 cjaMensuelle:str|None,
                 cjaJournaliere:str|None,
                 idCad:str,
                 nomTitulaire:str,
                 raisonSocialeTitulaire:str,
                 numeroSiretTitulaire:str,
                 dateMes:str|None,
                 dateMhs:str|None,
                 statutContractuel:str,
                 consommationJournalierePlafond:str|None,
                 modulationN1:str|None,
                 modulationN2:str|None,
                 modulationN3:str|None,
                 modulationN4:str|None,
                 assiette: str|None,
                 fournisseur:str,
                 profil:str,
                 dateDebutProfil:str,
                dateFinProfil:str


                 ):
        self.tarifAcheminement = tarifAcheminement
        self.carActuelle = carActuelle
        self.carFuture = carFuture
        self.profilTypeFutur = profilTypeFutur
        self.cja = cja
        self.cjaMensuelle = cjaMensuelle
        self.cjaJournaliere = cjaJournaliere
        self.idCad = idCad
        self.nomTitulaire = nomTitulaire
        self.raisonSocialeTitulaire = raisonSocialeTitulaire
        self.numeroSiretTitulaire = numeroSiretTitulaire
        self.dateMes = dateMes
        self.dateMhs = dateMhs
        self.statutContractuel = statutContractuel
        self.consommationJournalierePlafond = consommationJournalierePlafond
        self.modulationN1 = modulationN1
        self.modulationN2 = modulationN2
        self.modulationN3 = modulationN3
        self.modulationN4 = modulationN4
        self.assiette = assiette
        self.fournisseur = fournisseur
        self.profil = profil
        self.dateDebutProfil = dateDebutProfil
        self.dateFinProfil = dateFinProfil
        



class ConsommationReference:
    """Class representing a reference consommation from API"""
    def __init__(self,
                 id:int,
                 consommationType:str,
                 profile:str,
                 annee:int,
                 mois1:str,
                 mois2:str,
                 mois3:str,
                 mois4:str,
                 mois5:str,
                 mois6:str,
                 mois7:str,
                 mois8:str,
                 mois9:str,
                 mois10:str,
                 mois11:str,
                 mois12:str,
                 unite:str):
        self.id = id
        self.consommationType = consommationType
        self.profile = profile
        self.annee = annee
        self.mois1 = mois1
        self.mois2 = mois2
        self.mois3 = mois3
        self.mois4 = mois4
        self.mois5 = mois5
        self.mois6 = mois6
        self.mois7 = mois7
        self.mois8 = mois8
        self.mois9 = mois9
        self.mois10 = mois10
        self.mois11 = mois11
        self.mois12 = mois12
        self.unite = unite
    


"""Support for Exception custom."""
class ClientError(Exception):
    """Exception to indicate a general API error."""


class ClientCommunicationError(
    ClientError,
):
    """Exception to indicate a communication error."""


class ClientAuthenticationError(
    ClientError,
):
    """Exception to indicate an authentication error."""



import os
from pygazpar.datasource import TestDataSource, JsonFileDataSource, ExcelFileDataSource, JsonWebDataSource, ExcelWebDataSource
from pygazpar.enum import Frequency
from datetime import date, timedelta
from dotenv import load_dotenv


class TestAllDataSource:

    # ------------------------------------------------------
    @classmethod
    def setup_class(cls):
        """ setup any state specific to the execution of the given class (which
        usually contains tests).
        """

    # ------------------------------------------------------
    @classmethod
    def teardown_class(cls):
        """ teardown any state that was previously setup with a call to
        setup_class.
        """

    # ------------------------------------------------------
    def setup_method(self):
        """ setup any state tied to the execution of the given method in a
        class.  setup_method is invoked for every test method of a class.
        """
        tmpdir = os.path.normpath(f"{os.getcwd()}/tmp")

        # We create the tmp directory if not already exists.
        if not os.path.exists(tmpdir):
            os.mkdir(tmpdir)

        load_dotenv()

        self.__username = os.environ["GRDF_USERNAME"]
        self.__password = os.environ["GRDF_PASSWORD"]
        self.__pceIdentifier = os.environ["PCE_IDENTIFIER"]
        self.__tmp_directory = tmpdir

    # ------------------------------------------------------
    def teardown_method(self):
        """ teardown any state that was previously setup with a setup_method
        call.
        """

    # ------------------------------------------------------
    def test_sample(self):

        dataSource = TestDataSource()

        endDate = date.today()
        startDate = endDate + timedelta(days=-365)

        data = dataSource.load(self.__pceIdentifier, startDate, endDate)

        assert (len(data[Frequency.DAILY.value]) == 711)

        assert (len(data[Frequency.WEEKLY.value]) == 102)

        assert (len(data[Frequency.MONTHLY.value]) == 24)

        assert (len(data[Frequency.YEARLY.value]) == 2)

    # ------------------------------------------------------
    def test_jsonfile_sample(self):

        dataSource = JsonFileDataSource("tests/resources/donnees_informatives.json", "tests/resources/temperatures.json")

        endDate = date.today()
        startDate = endDate + timedelta(days=-365)

        data = dataSource.load(self.__pceIdentifier, startDate, endDate, [Frequency.DAILY, Frequency.WEEKLY, Frequency.MONTHLY, Frequency.YEARLY])

        assert (len(data[Frequency.DAILY.value]) == 1096)

        assert (len(data[Frequency.WEEKLY.value]) == 155)

        assert (len(data[Frequency.MONTHLY.value]) == 36)

        assert (len(data[Frequency.YEARLY.value]) == 3)

    # ------------------------------------------------------
    def test_daily_excelfile_sample(self):

        dataSource = ExcelFileDataSource("tests/resources/Donnees_informatives_PCE_DAILY.xlsx")

        endDate = date.today()
        startDate = endDate + timedelta(days=-365)

        data = dataSource.load(self.__pceIdentifier, startDate, endDate, [Frequency.DAILY])

        assert (len(data[Frequency.DAILY.value]) == 363)

    # ------------------------------------------------------
    def test_weekly_excelfile_sample(self):

        dataSource = ExcelFileDataSource("tests/resources/Donnees_informatives_PCE_WEEKLY.xlsx")

        endDate = date.today()
        startDate = endDate + timedelta(days=-365)

        data = dataSource.load(self.__pceIdentifier, startDate, endDate, [Frequency.WEEKLY])

        assert (len(data[Frequency.WEEKLY.value]) == 53)

    # ------------------------------------------------------
    def test_monthly_excelfile_sample(self):

        dataSource = ExcelFileDataSource("tests/resources/Donnees_informatives_PCE_MONTHLY.xlsx")

        endDate = date.today()
        startDate = endDate + timedelta(days=-365)

        data = dataSource.load(self.__pceIdentifier, startDate, endDate, [Frequency.MONTHLY])

        assert (len(data[Frequency.MONTHLY.value]) == 13)

    # ------------------------------------------------------
    def test_yearly_excelfile_sample(self):

        dataSource = ExcelFileDataSource("tests/resources/Donnees_informatives_PCE_DAILY.xlsx")

        endDate = date.today()
        startDate = endDate + timedelta(days=-365)

        data = dataSource.load(self.__pceIdentifier, startDate, endDate, [Frequency.YEARLY])

        assert (len(data[Frequency.YEARLY.value]) == 1)

    # ------------------------------------------------------
    def test_jsonweb(self):

        dataSource = JsonWebDataSource(self.__username, self.__password)

        endDate = date.today()
        startDate = endDate + timedelta(days=-365)

        data = dataSource.load(self.__pceIdentifier, startDate, endDate, [Frequency.DAILY, Frequency.WEEKLY, Frequency.MONTHLY, Frequency.YEARLY])

        assert (len(data[Frequency.DAILY.value]) > 0)

        assert (len(data[Frequency.WEEKLY.value]) >= 51 and len(data[Frequency.WEEKLY.value]) <= 54)

        assert (len(data[Frequency.MONTHLY.value]) >= 11 and len(data[Frequency.MONTHLY.value]) <= 13)

        assert (len(data[Frequency.YEARLY.value]) == 1)

    # ------------------------------------------------------
    def test_excelweb(self):

        dataSource = ExcelWebDataSource(self.__username, self.__password, self.__tmp_directory)

        endDate = date.today()
        startDate = endDate + timedelta(days=-365)

        data = dataSource.load(self.__pceIdentifier, startDate, endDate, [Frequency.DAILY, Frequency.WEEKLY, Frequency.MONTHLY, Frequency.YEARLY])

        assert (len(data[Frequency.DAILY.value]) > 0)

        assert (len(data[Frequency.WEEKLY.value]) >= 51 and len(data[Frequency.WEEKLY.value]) <= 54)

        assert (len(data[Frequency.MONTHLY.value]) >= 12 and len(data[Frequency.MONTHLY.value]) <= 13)

        assert (len(data[Frequency.YEARLY.value]) == 1)



"""Support for Json parser."""
import json
import logging
from typing import Any, List, Dict
from datetime import datetime
from pygazpar.enum import PropertyName
from pygazpar.types.ConsommationType import ConsommationType
from pygazpar.types.RelevesResultType import RelevesResultType

INPUT_DATE_FORMAT = "%Y-%m-%d"

OUTPUT_DATE_FORMAT = "%d/%m/%Y"

Logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------------------------
class JsonParser:
    """ Class Json parser."""

    # ------------------------------------------------------
    @staticmethod
    def parse_result(data: ConsommationType, temperatures: Dict[str, Any], pce_identifier: str) -> List[RelevesResultType]:
        """ parse Json to result."""

        res = []



        # Timestamp of the data.
        data_timestamp = datetime.now().isoformat()

        for releve in data.releves:
            temperature = releve.temperature
            if temperature is None and temperatures is not None and len(temperatures) > 0:
                temperature = temperatures.get(releve.journeeGaziere)
            item = RelevesResultType(datetime.strftime(datetime.strptime(releve.journeeGaziere, INPUT_DATE_FORMAT),
                                                       OUTPUT_DATE_FORMAT),data_timestamp,releve,temperature)
            res.append(item)

        Logger.debug("Daily data read successfully from Json")

        return res
 # ------------------------------------------------------
    @staticmethod
    def parse(json_str: str, temperature_str: str, pce_identifier: str) -> List[Dict[str, Any]]:
        """ parse Json to dictionnary."""
        res = []
        data = json.loads(json_str)

        temperatures = json.loads(temperature_str)

        # Timestamp of the data.
        data_timestamp = datetime.now().isoformat()

        for releve in data[pce_identifier]['releves']:
            temperature = releve['temperature']
            if temperature is None and temperatures is not None and len(temperatures) > 0:
                releve.temperature = temperatures.get(releve['journeeGaziere'])

            item = {}
            item[PropertyName.TIME_PERIOD.value] = datetime.strftime(
                datetime.strptime(releve['journeeGaziere'], INPUT_DATE_FORMAT),OUTPUT_DATE_FORMAT)
            item[PropertyName.START_INDEX.value] = releve['indexDebut']
            item[PropertyName.END_INDEX.value] = releve['indexFin']
            item[PropertyName.VOLUME.value] = releve['volumeBrutConsomme']
            item[PropertyName.ENERGY.value] = releve['energieConsomme']
            item[PropertyName.CONVERTER_FACTOR.value] = releve['coeffConversion']
            item[PropertyName.TEMPERATURE.value] = temperature
            item[PropertyName.QUALIFICATION.value] = releve['qualificationReleve']
            item[PropertyName.TIMESTAMP.value] = data_timestamp

            res.append(item)

        Logger.debug("Daily data read successfully from Json")
        return res



"""Support for GRDF authentication."""
from __future__ import annotations
import logging
import aiohttp
from yarl import URL
from .helpers import _api_wrapper
from .exceptions import ClientError

SESSION_TOKEN_URL = "https://connexion.grdf.fr/api/v1/authn"
AUTH_TOKEN_URL = "https://connexion.grdf.fr/login/sessionCookieRedirect"
LOG = logging.getLogger(__name__)

class GazparAuth:
    '''Manage the Auth API connection'''
    # ------------------------------------------------------
    def __init__(self, username: str, password: str, session: aiohttp.ClientSession, token: str = None):

        self.__username = username
        self.__password = password
        self._session = session
        self._token = token
    # ------------------------------------------------------
    async def request_token(self) -> str:
        '''Request the token to the API'''
        response= await _api_wrapper(
        session=self._session,
        method="post",
        url=SESSION_TOKEN_URL,
        headers={"Content-type": "application/json", "domain":"grdf.fr","X-Requested-With": "XMLHttpRequest"},
        data={"username": self.__username,"password": self.__password,"options":
              {"multiOptionalFactorEnroll": "false","warnBeforePasswordExpired": "false"}},
        )
        if response.content_type=="application/json":
            responsejson=await response.json()
            session_token = responsejson.get("sessionToken")
        else:
            raise ClientError("Invalid response from server")
        LOG.debug("Session token: %s", session_token)
        response=await _api_wrapper(
            session=self._session,
            method="get",
            url=AUTH_TOKEN_URL,
            headers={"Content-type": "application/json","X-Requested-With": "XMLHttpRequest"},
            params={"checkAccountSetupComplete": "true","token": session_token,"redirectUrl": "https://monespace.grdf.fr"},

        )
        auth_token = self._session.cookie_jar.filter_cookies(URL("https://monespace.grdf.fr")).get("auth_token")
        self._token = auth_token.value
        return auth_token.value  # type: ignore



"""Support for Datasource."""
from typing import Any, List, Dict, cast, Optional
import logging
import glob
import os
import json
import time
from datetime import date, timedelta
from abc import ABC, abstractmethod
import aiohttp
from pygazpar.enum import Frequency, PropertyName,ConsommationRole
from pygazpar.excelparser import ExcelParser
from pygazpar.jsonparser import JsonParser
from pygazpar.auth import GazparAuth
from pygazpar.consommation import GazparConsommation
from pygazpar.pce import GazparPCE
from pygazpar.frequency import FrequencyConverter
from pygazpar.types.PceType import PceType
Logger = logging.getLogger(__name__)

MeterReading = Dict[str, Any]

MeterReadings = List[MeterReading]

MeterReadingsByFrequency = Dict[str, MeterReadings]


# ------------------------------------------------------------------------------------------------------------
class IDataSource(ABC):
    '''Base class'''
    @abstractmethod
    async def load(self, pce_identifier: str, start_date: date, end_date: date, frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:
        '''Load data conso from source'''
        pass
    @abstractmethod
    async def login(self) -> str:
        '''Login from source'''
        pass

    @abstractmethod
    async def list_pce(self) -> List[PceType]:
        '''List PCE from source'''
        pass


# ------------------------------------------------------------------------------------------------------------
class WebDataSource(IDataSource):
    '''Base class for the WEB api'''
    # ------------------------------------------------------
    def __init__(self, username: str, password: str, session: aiohttp.ClientSession):

        self.__username = username
        self.__password = password
        self.__session = session
        self._pce= GazparPCE(session)
        self._conso=GazparConsommation(session)
        self._auth=GazparAuth(username, password,session)
        self._auth_token=None
    async def login(self) -> str:
         self._auth_token=await self._auth.request_token()
         return self._auth_token
    async def list_pce(self) -> List[PceType]:
         return await self._pce.get_list_pce()
    # ------------------------------------------------------
    async def load(self, pce_identifier: str, start_date: date, end_date: date, frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:

        if(self._auth_token is None):
            self._auth_token=await self._auth.request_token()
        
        res = await self._load_from_session(pce_identifier, start_date, end_date, frequencies)

        Logger.debug("The data update terminates normally")

        return res


    @abstractmethod
    async def _load_from_session(self, pce_identifier: str, start_date: date, end_date: date, frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:
        '''Load data from session'''
        pass


# ------------------------------------------------------------------------------------------------------------
class ExcelWebDataSource(WebDataSource):
    '''Base class for the Excel WEB api'''

    DATE_FORMAT = "%Y-%m-%d"

    FREQUENCY_VALUES = {
        Frequency.HOURLY: "Horaire",
        Frequency.DAILY: "Journalier",
        Frequency.WEEKLY: "Hebdomadaire",
        Frequency.MONTHLY: "Mensuel",
        Frequency.YEARLY: "Journalier"
    }

    DATA_FILENAME = 'Donnees_informatives_*.xlsx'

    # ------------------------------------------------------
    def __init__(self, username: str, password: str,tmpDirectory: str, session: aiohttp.ClientSession|None=None):

        if session is None:
            session = aiohttp.ClientSession(cookie_jar= aiohttp.CookieJar())
      
        super().__init__(username, password,session)
        
        self.__tmp_directory = tmpDirectory
    
    # ------------------------------------------------------
    async def _load_from_session(self, pce_identifier: str, start_date: date, end_date: date, frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:

        res = {}

        # XLSX is in the TMP directory
        data_file_path_pattern = self.__tmp_directory + '/' + ExcelWebDataSource.DATA_FILENAME

        # We remove an eventual existing data file (from a previous run that has not deleted it).
        file_list = glob.glob(data_file_path_pattern)
        for filename in file_list:
            if os.path.isfile(filename):
                try:
                    os.remove(filename)
                except PermissionError:
                    pass

        if frequencies is None:
            # Transform Enum in List.
            frequency_list = [frequency for frequency in Frequency]
        else:
            # Get unique values.
            frequency_list = set(frequencies)

        for frequency in frequency_list:
            # Inject parameters.

            Logger.debug(f"Loading data of frequency {ExcelWebDataSource.FREQUENCY_VALUES[frequency]} from {start_date.strftime(ExcelWebDataSource.DATE_FORMAT)} to {end_date.strftime(ExcelWebDataSource.DATE_FORMAT)}")

            # Retry mechanism.
            retry = 10
            while retry > 0:


                try:
                    response = await self._conso.get_consommation_file(pce_identifier,start_date.strftime(ExcelWebDataSource.DATE_FORMAT),end_date.strftime(ExcelWebDataSource.DATE_FORMAT),ConsommationRole.INFORMATIVES,frequency)
                    open(f"{self.__tmp_directory}/{response.filename}", "wb").write(response.content)

                    break
                except Exception as e:
                    if retry == 1:
                        raise e
                    Logger.error("An error occurred while loading data. Retry in 3 seconds.")
                    time.sleep(3)
                    retry -= 1

            # Load the XLSX file into the data structure
            file_list = glob.glob(data_file_path_pattern)

            if len(file_list) == 0:
                Logger.warning(f"Not any data file has been found in '{self.__tmp_directory}' directory")

            for filename in file_list:
                res[frequency.value] = ExcelParser.parse(filename, frequency if frequency != Frequency.YEARLY else Frequency.DAILY)
                try:
                    # openpyxl does not close the file properly.
                    os.remove(filename)
                except PermissionError:
                    pass

            # We compute yearly from daily data.
            if frequency == Frequency.YEARLY:
                res[frequency.value] = FrequencyConverter.compute_yearly(res[frequency.value])

        return res

   


# ------------------------------------------------------------------------------------------------------------
class ExcelFileDataSource(IDataSource):
    '''Base class for the Excel file'''
    def __init__(self, excel_file: str):

        self.__excel_file = excel_file
    async def login(self) -> str:
         pass
    async def list_pce(self) -> List[PceType]:
        '''List PCE from source'''
        pass
    async def load(self, pce_identifier: str, start_date: date,
                   end_date: date, frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:

        res = {}

        if frequencies is None:
            # Transform Enum in List.
            frequency_list = [frequency for frequency in Frequency]
        else:
            # Get unique values.
            frequency_list = set(frequencies)

        for frequency in frequency_list:
            if frequency != Frequency.YEARLY:
                res[frequency.value] = ExcelParser.parse(self.__excel_file, frequency)
            else:
                daily = ExcelParser.parse(self.__excel_file, Frequency.DAILY)
                res[frequency.value] = FrequencyConverter.compute_yearly(daily)

        return res


# ------------------------------------------------------------------------------------------------------------
class JsonWebDataSource(WebDataSource):
    '''Base class for the Json Web data'''
    INPUT_DATE_FORMAT = "%Y-%m-%d"
    OUTPUT_DATE_FORMAT = "%d/%m/%Y"

    def __init__(self, username: str, password: str, session: aiohttp.ClientSession|None=None):

        if session is None:
            session = aiohttp.ClientSession(cookie_jar= aiohttp.CookieJar())
        super().__init__(username, password,session)

    async def _load_from_session(self,pce_identifier: str, start_date: date, end_date: date, 
                                 frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:

        res = {}

        compute_by_frequency = {
            Frequency.HOURLY: FrequencyConverter.compute_hourly,
            Frequency.DAILY: FrequencyConverter.compute_daily,
            Frequency.WEEKLY: FrequencyConverter.compute_weekly,
            Frequency.MONTHLY: FrequencyConverter.compute_monthly,
            Frequency.YEARLY: FrequencyConverter.compute_yearly
        }

        # Data URL: Inject parameters.
        # Retry mechanism.
        retry = 10
        while retry > 0:


            try:
                data=await self._conso.get_consommation(pce_identifier,start_date.strftime(JsonWebDataSource.INPUT_DATE_FORMAT),
                                                        end_date.strftime(JsonWebDataSource.INPUT_DATE_FORMAT),ConsommationRole.INFORMATIVES)
                break
            except Exception as e:

                if retry == 1:
                    raise e

                Logger.error("An error occurred while loading data. Retry in 3 seconds.")
                time.sleep(3)
                retry -= 1

        # Temperatures URL: Inject parameters.
        end_date = date.today() - timedelta(days=1) if end_date >= date.today() else end_date
        days = min((end_date - start_date).days, 730)
        # Get weather data.
        temperatures=await self._pce.get_pce_meteo(pce_identifier,end_date.strftime(JsonWebDataSource.INPUT_DATE_FORMAT),days)

        # Transform all the data into the target structure.
        daily = JsonParser.parse_result(data, temperatures, pce_identifier)

        if frequencies is None:
            # Transform Enum in List.
            frequency_list = [frequency for frequency in Frequency]
        else:
            # Get unique values.
            frequency_list = set(frequencies)

        for frequency in frequency_list:
            res[frequency.value] = compute_by_frequency[frequency](daily)

        return res


# ------------------------------------------------------------------------------------------------------------
class JsonFileDataSource(IDataSource):
    '''Base class for the Json File data'''
    def __init__(self, consumption_json_file: str, temperature_json_file):

        self.__consumption_json_file = consumption_json_file
        self.__temperature_json_file = temperature_json_file
    async def login(self) -> str:
         pass
    async def list_pce(self) -> List[PceType]:
        '''List PCE from source'''
        pass
    async def load(self, pce_identifier: str, start_date: date, end_date: date,
                   frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:

        res = {}

        with open(self.__consumption_json_file) as __consumption_json_file:
            with open(self.__temperature_json_file) as __temperature_json_file:
                daily = JsonParser.parse(__consumption_json_file.read(), __temperature_json_file.read(), pce_identifier)

        compute_by_frequency = {
            Frequency.HOURLY: FrequencyConverter.compute_hourly,
            Frequency.DAILY: FrequencyConverter.compute_daily,
            Frequency.WEEKLY: FrequencyConverter.compute_weekly,
            Frequency.MONTHLY: FrequencyConverter.compute_monthly,
            Frequency.YEARLY: FrequencyConverter.compute_yearly
        }

        if frequencies is None:
            # Transform Enum in List.
            frequency_list = [frequency for frequency in Frequency]
        else:
            # Get unique values.
            frequency_list = set(frequencies)

        for frequency in frequency_list:
            res[frequency.value] = compute_by_frequency[frequency](daily)

        return res


# ------------------------------------------------------------------------------------------------------------
class TestDataSource(IDataSource):
    '''Base class for the Test data'''
    def __init__(self):
        pass
    async def login(self) -> str:
         pass
    async def list_pce(self) -> List[PceType]:
        '''List PCE from source'''
        pass
    async def load(self, pce_identifier: str, start_date: date, end_date: date,
                   frequencies: Optional[List[Frequency]] = None) -> MeterReadingsByFrequency:

        res = {}
        data_sample_filename_by_frequency = {
            Frequency.HOURLY: "hourly_data_sample.json",
            Frequency.DAILY: "daily_data_sample.json",
            Frequency.WEEKLY: "weekly_data_sample.json",
            Frequency.MONTHLY: "monthly_data_sample.json",
            Frequency.YEARLY: "yearly_data_sample.json"
        }

        if frequencies is None:
            # Transform Enum in List.
            frequency_list = [frequency for frequency in Frequency]
        else:
            # Get unique values.
            frequency_list = set(frequencies)

        for frequency in frequency_list:
            data_sample_filename = f"{os.path.dirname(os.path.abspath(__file__))}/resources/{data_sample_filename_by_frequency[frequency]}"

            with open(data_sample_filename) as json_file:
                res[frequency.value] = cast(List[Dict[PropertyName, Any]], json.load(json_file))

        return res



"""Support for Excel parser."""
from typing import  List, Dict
import logging
from datetime import datetime, time,timedelta
import pytz
import dateparser
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.cell.cell import Cell
from openpyxl import load_workbook
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from pygazpar.enum import NatureReleve, QualificationReleve, StatusReleve,Frequency,PropertyName
from pygazpar.types.RelevesResultType import RelevesResultType
from pygazpar.types.ConsommationType import RelevesType
FIRST_DATA_LINE_NUMBER = 10

Logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------------------------
class ExcelParser:
    '''Excel parser for releve'''
    OUTPUT_DATE_FORMAT = "%Y-%m-%d"
    OUTPUT_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S%z"

    INPUT_DATE_FORMAT = "%d/%m/%Y"
    # ------------------------------------------------------
    @staticmethod
    def parse(data_filename: str, data_reading_frequency: Frequency) -> List[RelevesResultType]:
        '''Parse excel file'''
        parse_by_frequency = {
            Frequency.HOURLY: ExcelParser.__parse_hourly,
            Frequency.DAILY: ExcelParser.__parse_daily,
            Frequency.WEEKLY: ExcelParser.__parse_weekly,
            Frequency.MONTHLY: ExcelParser.__parse_monthly
        }

        Logger.debug(f"Loading Excel data file '{data_filename}'...")

        workbook = load_workbook(filename=data_filename)

        worksheet = workbook.active

        res = parse_by_frequency[data_reading_frequency](worksheet)  # type: ignore

        workbook.close()

        return res

    # ------------------------------------------------------
    @staticmethod
    def __fill_row(row: Dict, property_name: str, cell: Cell, is_number: bool):
        '''fill row dictionnary with value from excel'''
        if cell.value is not None:
            if is_number:
                if isinstance(cell.value,str):
                    if len(cell.value.strip()) > 0:
                        row[property_name] = float(cell.value.replace(',', '.'))
                else:
                    row[property_name] = cell.value
            else:
                row[property_name] = cell.value.strip() if isinstance(cell.value,str) else cell.value
        else:
            row[property_name] = None

    # ------------------------------------------------------
    @staticmethod
    def __parse_hourly(worksheet: Worksheet) -> List[RelevesResultType]:
        '''Parse hourly data'''
        return []

    # ------------------------------------------------------
    @staticmethod
    def __parse_daily(worksheet: Worksheet) -> List[RelevesResultType]:
        '''parse daily data'''
        res = []       
        # Timestamp of the data.
        data_timestamp = datetime.now().isoformat()

        minRowNum = FIRST_DATA_LINE_NUMBER
        maxRowNum = len(worksheet['B'])
        for rownum in range(minRowNum, maxRowNum + 1):
            row = {}
            if worksheet.cell(column=2, row=rownum).value is not None:
                date_journee = datetime.strptime(worksheet.cell(column=2, row=rownum).value, ExcelParser.INPUT_DATE_FORMAT).date()
                info=pytz.timezone('Europe/Paris')
                MyTime = time(6, 0, 0)  #hr/min/sec
                datetime_debut = datetime.combine(date_journee, MyTime)
                datetime_debut_localize=info.localize(datetime_debut)
                row[PropertyName.JOURNEE_GAZIERE.value] = date_journee.strftime(ExcelParser.OUTPUT_DATE_FORMAT)
                row[PropertyName.DATE_DEBUT.value]= datetime_debut_localize.isoformat()
                row[PropertyName.DATE_FIN.value]= (datetime_debut_localize+timedelta(days=1)).isoformat()

                ExcelParser.__fill_row(row, PropertyName.START_INDEX.value, worksheet.cell(column=3, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.END_INDEX.value, worksheet.cell(column=4, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.VOLUME.value, worksheet.cell(column=5, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.ENERGY.value, worksheet.cell(column=6, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.CONVERTER_FACTOR.value, worksheet.cell(column=7, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.TEMPERATURE.value, worksheet.cell(column=8, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.QUALIFICATION.value, worksheet.cell(column=9, row=rownum), False)  # type: ignore
                row[PropertyName.PCS.value]=None
                row[PropertyName.VOLUME_CONVERTI.value]=round(row[PropertyName.VOLUME.value])
                row[PropertyName.PTA.value]=None
                row[PropertyName.NATURE.value]=NatureReleve.INFORMATIVES.value
                row[PropertyName.STATUS.value]=StatusReleve.PROVISOIRE.value
                row[PropertyName.FREQUENCE_RELEVE.value]=None
                releve = RelevesType(**row)
                releve_result = RelevesResultType(worksheet.cell(column=2, row=rownum).value,data_timestamp,releve)

                res.append(releve_result)

        Logger.debug(f"Daily data read successfully between row #{minRowNum} and row #{maxRowNum}")

        return res

    # ------------------------------------------------------
    @staticmethod
    def __parse_weekly(worksheet: Worksheet) -> List[RelevesResultType]:
        '''parse weekly data'''
        res = []

        # Timestamp of the data.
        data_timestamp = datetime.now().isoformat()
        info=pytz.timezone('Europe/Paris')
        my_time = time(6, 0, 0)  #hr/min/sec
        min_row_num = FIRST_DATA_LINE_NUMBER
        max_row_num = len(worksheet['B'])
        for rownum in range(min_row_num, max_row_num + 1):
            row = {}
            if worksheet.cell(column=2, row=rownum).value is not None:
                dateField=worksheet.cell(column=2, row=rownum).value
                dateStart=dateField.split('au')[0]
                dateEnd=dateField.split('au')[1]
                dateStartDT=parse(dateStart, fuzzy_with_tokens=True)
                dateEndDT=parse(dateEnd, fuzzy_with_tokens=True)
                dateStartDT = datetime.combine(dateStartDT[0], my_time)
                dateEndDT = datetime.combine(dateEndDT[0], my_time)
                dateStartDT=info.localize(dateStartDT)
                dateEndDT=info.localize(dateEndDT)

                row[PropertyName.DATE_DEBUT.value]= dateStartDT.isoformat()
                row[PropertyName.DATE_FIN.value]= (dateEndDT+timedelta(days=1)).isoformat()
                row[PropertyName.JOURNEE_GAZIERE.value] =None
                ExcelParser.__fill_row(row, PropertyName.VOLUME.value, worksheet.cell(column=3, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.ENERGY.value, worksheet.cell(column=4, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.TEMPERATURE.value, worksheet.cell(column=5, row=rownum), True)  # type: ignore

                row[PropertyName.START_INDEX.value]=None
                row[PropertyName.END_INDEX.value]=None
                row[PropertyName.CONVERTER_FACTOR.value]=None
                row[PropertyName.QUALIFICATION.value]=QualificationReleve.ESTIME.value

                row[PropertyName.PCS.value]=None
                row[PropertyName.VOLUME_CONVERTI.value]=round(row[PropertyName.VOLUME.value])
                row[PropertyName.PTA.value]=None
                row[PropertyName.NATURE.value]=NatureReleve.INFORMATIVES.value
                row[PropertyName.STATUS.value]=StatusReleve.PROVISOIRE.value
                row[PropertyName.FREQUENCE_RELEVE.value]=None
                releve = RelevesType(**row)
                releve_result = RelevesResultType(worksheet.cell(column=2, row=rownum).value,data_timestamp,releve)
                res.append(releve_result)

        Logger.debug(f"Weekly data read successfully between row #{min_row_num} and row #{max_row_num}")

        return res

    # ------------------------------------------------------
    @staticmethod
    def __parse_monthly(worksheet: Worksheet) -> List[RelevesResultType]:
        '''parse Monthly data'''
        res = []

        # Timestamp of the data.
        data_timestamp = datetime.now().isoformat()
        info=pytz.timezone('Europe/Paris')
       
        MyTime = time(6, 0, 0)  #hr/min/sec
        minRowNum = FIRST_DATA_LINE_NUMBER
        maxRowNum = len(worksheet['B'])
        for rownum in range(minRowNum, maxRowNum + 1):
            row = {}
            if worksheet.cell(column=2, row=rownum).value is not None:
                dateField=worksheet.cell(column=2, row=rownum).value
                dateStartDT=dateparser.parse(dateField, locales=['fr'])
                dateStartDT=dateStartDT.replace(day=1)
                dateStartDT = datetime.combine(dateStartDT, MyTime)
                dateStartDT=info.localize(dateStartDT)
                row[PropertyName.DATE_DEBUT.value]= dateStartDT.isoformat()
                row[PropertyName.DATE_FIN.value]= (dateStartDT+relativedelta(months=1)).isoformat()
                row[PropertyName.JOURNEE_GAZIERE.value] =None

                ExcelParser.__fill_row(row, PropertyName.VOLUME.value, worksheet.cell(column=3, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.ENERGY.value, worksheet.cell(column=4, row=rownum), True)  # type: ignore
                ExcelParser.__fill_row(row, PropertyName.TEMPERATURE.value, worksheet.cell(column=5, row=rownum), True)  # type: ignore

                row[PropertyName.START_INDEX.value]=None
                row[PropertyName.END_INDEX.value]=None
                row[PropertyName.CONVERTER_FACTOR.value]=None
                row[PropertyName.QUALIFICATION.value]=QualificationReleve.ESTIME.value
                row[PropertyName.PCS.value]=None
                row[PropertyName.VOLUME_CONVERTI.value]=round(row[PropertyName.VOLUME.value])
                row[PropertyName.PTA.value]=None
                row[PropertyName.NATURE.value]=NatureReleve.INFORMATIVES.value
                row[PropertyName.STATUS.value]=StatusReleve.PROVISOIRE.value
                row[PropertyName.FREQUENCE_RELEVE.value]=None
                
                releve = RelevesType(**row)
                releve_result = RelevesResultType(worksheet.cell(column=2, row=rownum).value,data_timestamp,releve)
                res.append(releve_result)

        Logger.debug(f"Monthly data read successfully between row #{minRowNum} and row #{maxRowNum}")

        return res



from pygazpar.types.DetailsPceType import DetailsPce
from pygazpar.types.ContratType import ContratPce
class PceType:
    """Class representing a PCE from API"""
    def __init__(self, 
                idObject:str, 
                typeObject:str|None,
                role:str,
                alias:str,
                teleReleve:bool,
                pce:str,
                dateActivation:str,
                dateMhs:str|None,
                dateMes:str|None,
                codePostal:str,
                frequenceReleve:str,
                etat:str,
                datePremiereAccreditation:str,
                nomTitulaire:str,
                 idAccreditation:str|None,
                raisonSociale:str|None,
                denominationClient:str|None,
                adresseEmailClient:str|None,
                telephoneClient:str|None,
                dateCreation:str|None,
                dateDebutConsentement:str|None,
                dateFinConsentement:str|None,
                dateDebutAccesDonneesConso:str|None,
                dateFinAccesDonneesConso:str|None,
                dateEtat:str|None,
                donneesConsoPubliees:str|None,
                donneesConsoInformatives:str|None,
                donneesContractuelles:str|None,
                donneesTechniques:str|None,
                parcours: str|None,
                statutControlePreuves:str|None,
                dateLimitePreuves:str|None,
                details:DetailsPce|None,
                dateDerniereVerification:str,
                frequenceJJ:bool = None,
                frequence1:bool =None,
                frequenceMM:bool =None,
                frequence6M:bool =None,
                frequenceMMOrJJ:bool =None,
                numeroSerie:str =None,
                numeroMatricule:str =None,
                contrat:ContratPce=None,
                fullAddress:str=None,
                **kwargs
                 ):
        self.idObject = idObject
        self.typeObject = typeObject
        self.idAccreditation = idAccreditation
        self.raisonSociale = raisonSociale
        self.denominationClient = denominationClient
        self.adresseEmailClient = adresseEmailClient
        self.telephoneClient = telephoneClient
        self.dateCreation = dateCreation
        self.dateDebutConsentement = dateDebutConsentement
        self.dateFinConsentement = dateFinConsentement
        self.dateDebutAccesDonneesConso = dateDebutAccesDonneesConso
        self.dateFinAccesDonneesConso = dateFinAccesDonneesConso
        self.dateEtat = dateEtat
        self.donneesConsoPubliees = donneesConsoPubliees
        self.donneesConsoInformatives = donneesConsoInformatives
        self.donneesContractuelles = donneesContractuelles
        self.donneesTechniques = donneesTechniques
        self.parcours = parcours
        self.statutControlePreuves = statutControlePreuves
        self.dateLimitePreuves = dateLimitePreuves
        self.role =role
        self.alias = alias
        self.teleReleve = teleReleve
        self.pce = pce
        self.dateActivation = dateActivation
        self.dateMhs = dateMhs
        self.dateMes = dateMes
        self.codePostal = codePostal
        self.frequenceReleve = frequenceReleve
        self.etat = etat
        self.datePremiereAccreditation = datePremiereAccreditation
        self.nomTitulaire = nomTitulaire
        if not isinstance(details, DetailsPce) and isinstance(details, dict):
            self.details=DetailsPce(**details)
        else:
            self.details = details
      
        self.dateDerniereVerification = dateDerniereVerification
        self.frequenceJJ = frequenceJJ
        self.frequence1 = frequence1
        self.frequenceMM = frequenceMM
        self.frequence6M = frequence6M
        self.frequenceMMOrJJ = frequenceMMOrJJ
        self.numeroSerie = numeroSerie
        self.numeroMatricule = numeroMatricule
        if not isinstance(contrat, ContratPce) and isinstance(contrat, dict):
            self.contrat=ContratPce(**contrat)
        else:
            self.contrat = contrat
        
        self.fullAddress = fullAddress